/*
 * File:   ecmfactorizator.cpp
 * Copyright 2009 Filippo Bergamasco (fbergama@gmail.com)
 *
 * This file is part of infact.
 *
 * infact is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * infact is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with infact.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created on January 18, 2009, 1:52 PM
 */

#include "includes/algorithms/ecmfactorizator.h"
#include "includes/infact.h"
#include "includes/mpnumbergenerator.h"



ECMFactorizator::~ECMFactorizator() {
}


FactorizationAlgorithm* ECMFactorizator::newInstance( mpz_class* n ) {
    if( infact::vv )
        std::cout <<  "Created new instance of ECMFactorizator" << std::endl;
    return new ECMFactorizator( n );
}


/*
 * Let P=(x1,y1) in E(Zn)
 * Let Q=(x2,y2) in E(Zn)
 *
 * if P+Q not equal O (point at infinity) ecm_sum
 * returns true and
 * R = (xr,yr) = P+Q
 *
 * if P+Q equals 0 ecm_sum returns false and
 * gcd = GCD( x2-x1, n )
 *
 */
inline bool ecm_sum( mpz_class* gcd,
                     mpz_class* xr, mpz_class* yr,
                     mpz_class* x1, mpz_class* y1,
                     mpz_class* x2, mpz_class* y2, mpz_class* n )
{
   mpz_class x2subx1;
   x2subx1 = (*x2)-(*x1);

   mpz_class x2subx1inv;
   mpz_class dummy;

   // Calc x2subx1inv if it exists
   mpz_gcdext( gcd->get_mpz_t(), x2subx1inv.get_mpz_t(), dummy.get_mpz_t(), x2subx1.get_mpz_t(), n->get_mpz_t() );

   if( *gcd != 1 ) {
       return false;
   }

   // Calc xr and yr
   mpz_class lambda = ( ( (*y2) - (*y1) )*x2subx1inv );
   (*xr) = (lambda*lambda) - (*x1) - (*x2);
   mpz_mod( xr->get_mpz_t(), xr->get_mpz_t(), n->get_mpz_t() );

   (*yr) = lambda*( (*x1)-(*xr) ) - (*y1);
   mpz_mod( yr->get_mpz_t(), yr->get_mpz_t(), n->get_mpz_t() );

   return true;
}


/*
 * Let P=(x1,y1) in E(Zn)
 *
 * if P+P not equal O (point at infinity) ecm_double
 * returns true and
 * R = (xr,yr) = P+P
 *
 * if P+P equals 0 ecm_sum returns false and
 * gcd = GCD( 2*y1, n )
 *
 */
inline bool ecm_double( mpz_class* gcd,
                        mpz_class* xr, mpz_class* yr,
                        mpz_class* x1, mpz_class* y1,
                        mpz_class* a, mpz_class* n )
{
   mpz_class dy1 = (*y1)*2;
   mpz_class dy1inv;
   mpz_class dummy;

   // Calc 2y1inv if it exists
   mpz_gcdext( gcd->get_mpz_t(), dy1inv.get_mpz_t(), dummy.get_mpz_t(), dy1.get_mpz_t(), n->get_mpz_t() );

   if( *gcd != 1 ) {
       return false;
   }
   
   // Calc xr and yr
   mpz_class lambda = (( ( ((*x1)*(*x1))*3 ) + (*a) )*dy1inv);
   
   (*xr) = (lambda*lambda) - ( (*x1)*2 ) ;
   mpz_mod( xr->get_mpz_t(), xr->get_mpz_t(), n->get_mpz_t() );

   (*yr) = ( lambda*( (*x1)-(*xr) ) ) - (*y1);
   mpz_mod( yr->get_mpz_t(), yr->get_mpz_t(), n->get_mpz_t() );

   return true;
}


mpz_class* ECMFactorizator::findNonTrivialFactor() {

    // Pre-conditions check
    if( mpz_probab_prime_p(n->get_mpz_t(), 20) ) {
        throw new infact::PrimeFactorizationException();
    }

    mpz_class* two = new mpz_class(2);
    if( mpz_divisible_p( n->get_mpz_t(), two->get_mpz_t() ) ) {
        return two;
    }
    delete two;


    char progressStr[] = {'-','\\','|','/'};
    char p = 0;
    short pn = 0;
    
    mpz_class B( n->get_mpz_t() );
    mpz_class max( n->get_mpz_t() );

    mpz_class* d = new mpz_class( n->get_mpz_t() );
    
    //Curve: y^2 = x^3 + ax + b
    MPNumberGenerator gen;
    mpz_class* r = gen.generateRand( n );
    (*r)+=2;
    mpz_class a( (*r) );
    delete r;

    if( infact::vv ) {
        std::cout << "Trying factorization with the following elliptic curve: " << std::endl;
        std::cout << "y^2 = x^3 + " << a.get_str(10) << "x + 1 " << std::endl;
    }

    
    B=2;
    mpz_class bfact(1);

    while( B < max ) {
        if( infact::vv )
            std::cout << "ECM factorization with bound B=" << B.get_str(10) << std::endl;
        
        //P=(x1,y1)
        mpz_class px(0);
        mpz_class py(1);
        mpz_realloc2( px.get_mpz_t(), 4096 );
        mpz_realloc2( py.get_mpz_t(), 4096 );
        mpz_class x1(px);
        mpz_class y1(py);
        mpz_class xnext;
        mpz_class ynext;

        if( !ecm_double(d, &xnext, &ynext, &x1, &y1, &a, n) ) {
            delete d;
            std::cout << "GCD != 1 at first step!! " << d->get_str(10) << std::endl;
            throw new infact::FactorizationFailureException("Strange error, GCD!=1 at first step");
        }
        x1=xnext;
        y1=ynext;

        bfact *= B;
        mpz_class i( bfact );

        while( i > 0 ) {
            if( !ecm_sum( d, &xnext, &ynext, &x1, &y1, &px, &py, n ) ) {
                
                if( (*d) != 1 && (*d)!=(*n) ) {
                    if( infact::progress)
                        std::cout << CLEAR_LINE;
                    return d;
                }
                delete d;
                throw new infact::FactorizationFailureException("ECM failed. Try again.");
            }
            
            x1=xnext;
            y1=ynext;
            i--;

            if( p<0 && infact::progress ) {
                std::cout << "\rProgress: " << progressStr[pn];
                pn++;
                pn%=4;
                p=0;
            }
            p++;
        }
        if( infact::progress)
            std::cout << CLEAR_LINE;
        B++;
    }

    delete d;
    throw new infact::FactorizationFailureException("ECM failed.");
}


std::string ECMFactorizator::getName() {
    return "ecm";
}
