/* 
 * File:   main.cpp
 * Copyright 2009 Filippo Bergamasco (fbergama@gmail.com)
 *
 * This file is part of infact.
 *
 * infact is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * infact is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with infact.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created on January 16, 2009, 3:25 PM
 *
 */

#include <string>


#include <boost/program_options/value_semantic.hpp>

#include "includes/infact.h"
#include "includes/mpnumbergenerator.h"
#include "includes/factorizationalgorithmfactory.h"
#include "includes/highrestimer.h"
#include <stdlib.h>
#include <iostream>

#include <boost/program_options/options_description.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>


namespace po = boost::program_options;
namespace infact {
    bool v=false;
    bool vv=false;
    bool progress=false;
}

int completeFactorization( FactorizationAlgorithm* alg ) {
    try {
        mpz_class* s = alg->findNonTrivialFactor();
        mpz_class* q = new mpz_class( alg->getN()->get_mpz_t() );
        mpz_div( q->get_mpz_t(), alg->getN()->get_mpz_t(), s->get_mpz_t());
        
        FactorizationAlgorithm* algS = alg->newInstance( s );
        FactorizationAlgorithm* algQ = alg->newInstance( q );
        int retCode = completeFactorization( algS ) || completeFactorization( algQ );
        delete alg;
        return retCode;
        
    } catch( infact::PrimeFactorizationException* pfe ) {
        std::cout << alg->getN()->get_str(10) << std::endl;
        delete alg;
        delete pfe;
        return(EXIT_SUCCESS);

    } catch( infact::FactorizationFailureException* ffe ) {
        std::cout << "Factorization failed: " << ffe->getDesc() << std::endl;
        delete alg;
        delete ffe;
        return(EXIT_SUCCESS);
    }
}

int findFactor( FactorizationAlgorithm* alg ) {
    try {
        mpz_class* s = alg->findNonTrivialFactor();
        std::cout << s->get_str(10) << std::endl;
        delete s;
        delete alg;
        return (EXIT_SUCCESS);

    } catch( infact::PrimeFactorizationException* pfe ) {
        std::cout << "Number is prime." << std::endl;
        delete alg;
        delete pfe;
        return(EXIT_SUCCESS);

    } catch( infact::FactorizationFailureException* ffe ) {
        std::cout << "Factorization failed: " << ffe->getDesc() << std::endl;
        delete alg;
        delete ffe;
        return(EXIT_SUCCESS);
    }
}


int benchmark( mpz_class* n, FactorizationAlgorithmFactory* factory ) {
    std::cout << "Benchmark mode. Number: " << n->get_str(10) << " (" <<
    n->get_str(10).length() << " digits)" <<std::endl;
    std::cout << "Algorithm,Time elapsed (10^-6 sec)" << std::endl;

    HighResTimer t;

    std::vector<std::string>* algNames = factory->getAlgorithmsNames();

    for( std::vector< std::string >::iterator it = algNames->begin(); it<algNames->end(); it++ ) {
        
        FactorizationAlgorithm* alg = factory->newInstance( n, (*it) );
        if( alg == NULL ) {
            std::cerr << "Error: Factorization algorithm not available." << std::endl;
            delete n;
            delete algNames;
            return -2;
        }

        try {
            t.reset();
            
            try {
                t.start();
                mpz_class* f = alg->findNonTrivialFactor();
                t.stop();
                delete f;
            }catch(infact::PrimeFactorizationException* pfe) {
                delete pfe;
            }
            
            std::cout << alg->getName() << "," << t.getMicroSecondsElapsed() << std::endl;
        } catch( infact::FactorizationFailureException* ffe ) {
            std::cout << alg->getName() << "," << "Fail." << std::endl;
        }

    }

    delete algNames;
    return EXIT_SUCCESS;
}

/*
 * 
 */
int main(int argc, char** argv) {

    unsigned int primesize;
    unsigned int compositesize;
    unsigned int attempts;
    std::string algName;
    bool complete=false;
    FactorizationAlgorithmFactory factory;

    //Program options initialization
    po::options_description desc("Program options");

    desc.add_options()
        ("help,h", "produce this help message\n")
        ("verbose,v", "produce debug messages\n")
        ("veryverbose,V", "produce a lot of debug messages.\n")
        ("algorithm,a", po::value< std::string >(&algName)->default_value( "trial-division"), "factorization algorithm to be used.\n"
                                                     )
        ("list-algorithms,l", "show a list of available algorithms and exit.\n")
        ("show-progress,s", "output progress status during computation.\n")
        ("complete,f", "complete factorization.\n"
                       "Infact will enumerate all prime factors of n.\n")
        ("benchmark,b", "Factorize with all algorithms and output data on elapsed time.\n")
        ("primegen,p", po::value< unsigned int >(&primesize), "generate a n-bits prime and exit.\n"
                                                              "prime generation is performed using Rabin-Miller "
                                                              "primality test.\n")
        ("compositegen,c", po::value< unsigned int >(&compositesize), "generate a n-bits composite and exit.\n")
        ("twoprimescompositegen,C", po::value< unsigned int >(&primesize), "generate a composite using two n-bits primes and exit.\n")
        ("attempts,r", po::value< unsigned int >(&attempts)->default_value(1000), "max number of trials before failure (for prime generation)\n")
    ;

    po::variables_map vm;
   
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    }catch( ... ) {
        std::cout << "Wrong arguments. See infact --help to get an overview of accepted commands." << std::endl;
        return -4;
    }

    if ( vm.count("help")) {
        std::cout << std::endl << "infact" << " v. " << _PROGRAM_VERSION_STR_ << std::endl;
        std::cout << "Filippo Bergamasco (fbergama@gmail.com) " << "\nUniversità Ca'Foscari di Venezia - d.s.i." << std::endl;
        std::cout << "------------------------" << std::endl << std::endl;
        std::cout << "Multi-precision INtegerFACTorizator." << std::endl << std::endl;
        std::cout << "Usage: " << std::endl;
        std::cout << "infact [options] < number  > number" << std::endl << std::endl;
        std::cout << "Input can be any lenght base-10 number string." << std::endl;
        std::cout << "Output: A non trivial factor of input number." << std::endl << std::endl;
        std::cout << desc << "\n";
        return 1;
    }

   
    infact::v =  vm.count("verbose");
    if( vm.count("veryverbose") )
        infact::v = infact::vv = true;

    complete = vm.count("complete");
    infact::progress = vm.count("show-progress");

    
    if( vm.count("list-algorithms") ) {
        std::cout << factory.availableAlgorithmsDescriptionString() << std::endl;
        return (EXIT_SUCCESS);
    }
    
    if( vm.count("primegen") ) {
        if( infact::vv )
            std::cout << "Args: " << primesize << "-bit prime generation requested" << std::endl;
        
        MPNumberGenerator numgen;
        mpz_class* p = numgen.generatePrime( primesize, attempts );
        if( p == NULL ) {
            std::cerr << "Prime generation failure." << std::endl;
            return -1;
        }
        std::cout << p->get_str(10) << std::endl;
        delete p;
        return (EXIT_SUCCESS);
    }

    if( vm.count("compositegen") ) {
        if( infact::vv )
            std::cout << "Args: " << compositesize << "-bit composite generation requested" << std::endl;

        MPNumberGenerator numgen;
        mpz_class* c = numgen.generateComposite( compositesize, attempts );
        if(c == NULL ) {
            std::cerr << "Composite generation failure." << std::endl;
            return -1;
        }
        std::cout << c->get_str(10) << std::endl;
        delete c;
        return (EXIT_SUCCESS);
    }

    if( vm.count("twoprimescompositegen") ) {
        if( infact::vv )
            std::cout << "Args: " << "Composite generation from two " << primesize << "-bit primes requested" << std::endl;

        MPNumberGenerator numgen;
        mpz_class* c = numgen.generate2pComposite( primesize, attempts );
        if( c == NULL ) {
            std::cerr << "Composite generation failure." << std::endl;
            return -1;
        }
        std::cout << c->get_str(10) << std::endl;
        delete c;
        return (EXIT_SUCCESS);
    }


    
    
    if( infact::vv ) {
        std::cout << "Reading n from standard input " << std::endl;
    }

    std::string nstr;
    std::cin >> nstr;
    mpz_class* n;
    
    try {
        n = new mpz_class( nstr );
    } catch( std::invalid_argument e ) {
        std::cerr << "Fatal: Invalid number format." << std::endl;
        return -3;
    }

    if( vm.count("benchmark") ) {
        return benchmark( n, &factory );
    }

    // Factorization
    if( infact::vv ) {
        std::cout << "Instantiating " << algName << "... " << std::endl;
    }

    FactorizationAlgorithm* alg = factory.newInstance( n, algName );
    if( alg == NULL ) {
        std::cerr << "Factorization algorithm not available." << std::endl;
        delete n;
        return -2;
    }

    
    if( complete ) {
        if( infact::v )
            std::cout << "Starting complete factorization using " << algName << " algorithm." << std::endl;
        return completeFactorization( alg );
        
    } else {
        if( infact::v ) 
            std::cout << "Finding non trivial factor using " << algName << " algorithm." << std::endl;

        return findFactor( alg );
    }
    
    
    return (EXIT_SUCCESS);
}

